/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2013, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.cpp
    \brief    prediction class
*/

#include <memory.h>
#include "TComPrediction.h"

//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Constructor / destructor / initialize
// ====================================================================================================================

TComPrediction::TComPrediction()
: m_pLumaRecBuffer(0)
, m_iLumaRecStride(0)
{
  m_piYuvExt = NULL;
#if ILR_TM_ENH
  memset(m_pEnhPred, 0, sizeof(m_pEnhPred));
  m_iSRNum[0] = m_iSRNum[1] = 0;
#if ILR_TM_ENH_EXPLICIT_MODE
  m_pchEILRBlockMode = NULL;  
#endif
#if DEC_EILR_BLK
  memset(m_acYuvPred4ELR, 0, sizeof(m_acYuvPred4ELR));
#endif
#endif
}

TComPrediction::~TComPrediction()
{
  
  delete[] m_piYuvExt;

  m_acYuvPred[0].destroy();
  m_acYuvPred[1].destroy();

  m_cYuvPredTemp.destroy();

  if( m_pLumaRecBuffer )
  {
    delete [] m_pLumaRecBuffer;
  }
  
  Int i, j;
  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
    {
      m_filteredBlock[i][j].destroy();
    }
    m_filteredBlockTmp[i].destroy();
  }
#if ILR_TM_ENH 
  if(m_pEnhPred[0])
  {
    m_pEnhPred[0]->destroy();
    delete m_pEnhPred[0];
    m_pEnhPred[0] = NULL;
  }
  if(m_pEnhPred[1])
  {
    m_pEnhPred[1]->destroy();
    delete m_pEnhPred[1];
    m_pEnhPred[1] = NULL;
  }

#if ILR_TM_ENH_EXPLICIT_MODE
  if(m_pchEILRBlockMode)
  {
    if(m_pchEILRBlockMode[0])
    {
      xFree(m_pchEILRBlockMode[0]);
      m_pchEILRBlockMode[0] = NULL;
    }
    free(m_pchEILRBlockMode);
    m_pchEILRBlockMode = NULL;
  }
#endif

#if DEC_EILR_BLK
  if(m_acYuvPred4ELR[0])
  {
    m_acYuvPred4ELR[0]->destroy();
    delete m_acYuvPred4ELR[0];
    m_acYuvPred4ELR[0] = NULL;
  }
  if(m_acYuvPred4ELR[1])
  {
    m_acYuvPred4ELR[1]->destroy();
    delete m_acYuvPred4ELR[1];
    m_acYuvPred4ELR[1] = NULL;
  }
#endif

#endif
}

#if ILR_TM_ENH 
Void TComPrediction::initTempBuff(Bool bEnhPred, Bool bEnhPred4BlkDec)
#else
Void TComPrediction::initTempBuff()
#endif
{
  if( m_piYuvExt == NULL )
  {
    Int extWidth  = MAX_CU_SIZE + 16; 
    Int extHeight = MAX_CU_SIZE + 1;
    Int i, j;
    for (i = 0; i < 4; i++)
    {
      m_filteredBlockTmp[i].create(extWidth, extHeight + 7);
      for (j = 0; j < 4; j++)
      {
        m_filteredBlock[i][j].create(extWidth, extHeight);
      }
    }
    m_iYuvExtHeight  = ((MAX_CU_SIZE + 2) << 4);
    m_iYuvExtStride = ((MAX_CU_SIZE  + 8) << 4);
    m_piYuvExt = new Int[ m_iYuvExtStride * m_iYuvExtHeight ];

    // new structure
    m_acYuvPred[0] .create( MAX_CU_SIZE, MAX_CU_SIZE );
    m_acYuvPred[1] .create( MAX_CU_SIZE, MAX_CU_SIZE );

    m_cYuvPredTemp.create( MAX_CU_SIZE, MAX_CU_SIZE );

#if ILR_TM_ENH 
  if(bEnhPred)
  {
    m_pEnhPred[0] = new TComYuv;    
    m_pEnhPred[0]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
    m_pEnhPred[1] = new TComYuv;
    m_pEnhPred[1]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
  }
#if DEC_EILR_BLK
  if(bEnhPred4BlkDec)
  {
    m_acYuvPred4ELR[0] = new TComYuv;    
    m_acYuvPred4ELR[0]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
    m_acYuvPred4ELR[1] = new TComYuv;
    m_acYuvPred4ELR[1]->create(g_uiMaxCUWidth, g_uiMaxCUHeight);
  }
#endif
#endif
  }

  if (m_iLumaRecStride != (MAX_CU_SIZE>>1) + 1)
  {
    m_iLumaRecStride =  (MAX_CU_SIZE>>1) + 1;
    if (!m_pLumaRecBuffer)
    {
      m_pLumaRecBuffer = new Pel[ m_iLumaRecStride * m_iLumaRecStride ];
    }
  }
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

// Function for calculating DC value of the reference samples used in Intra prediction
Pel TComPrediction::predIntraGetPredValDC( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft )
{
  Int iInd, iSum = 0;
  Pel pDcVal;

  if (bAbove)
  {
    for (iInd = 0;iInd < iWidth;iInd++)
    {
      iSum += pSrc[iInd-iSrcStride];
    }
  }
  if (bLeft)
  {
    for (iInd = 0;iInd < iHeight;iInd++)
    {
      iSum += pSrc[iInd*iSrcStride-1];
    }
  }

  if (bAbove && bLeft)
  {
    pDcVal = (iSum + iWidth) / (iWidth + iHeight);
  }
  else if (bAbove)
  {
    pDcVal = (iSum + iWidth/2) / iWidth;
  }
  else if (bLeft)
  {
    pDcVal = (iSum + iHeight/2) / iHeight;
  }
  else
  {
    pDcVal = pSrc[-1]; // Default DC value already calculated and placed in the prediction array if no neighbors are available
  }
  
  return pDcVal;
}

// Function for deriving the angular Intra predictions

/** Function for deriving the simplified angular intra predictions.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 * \param dirMode the intra prediction mode index
 * \param blkAboveAvailable boolean indication if the block above is available
 * \param blkLeftAvailable boolean indication if the block to the left is available
 *
 * This function derives the prediction samples for the angular mode based on the prediction direction indicated by
 * the prediction mode index. The prediction direction is given by the displacement of the bottom row of the block and
 * the reference row above the block in the case of vertical prediction or displacement of the rightmost column
 * of the block and reference column left from the block in the case of the horizontal prediction. The displacement
 * is signalled at 1/32 pixel accuracy. When projection of the predicted pixel falls inbetween reference samples,
 * the predicted value for the pixel is linearly interpolated from the reference samples. All reference samples are taken
 * from the extended main reference.
 */
Void TComPrediction::xPredIntraAng(Int bitDepth, Int* pSrc, Int srcStride, Pel*& rpDst, Int dstStride, UInt width, UInt height, UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable, Bool bFilter )
{
  Int k,l;
  Int blkSize        = width;
  Pel* pDst          = rpDst;

  // Map the mode index to main prediction direction and angle
  assert( dirMode > 0 ); //no planar
  Bool modeDC        = dirMode < 2;
  Bool modeHor       = !modeDC && (dirMode < 18);
  Bool modeVer       = !modeDC && !modeHor;
  Int intraPredAngle = modeVer ? (Int)dirMode - VER_IDX : modeHor ? -((Int)dirMode - HOR_IDX) : 0;
  Int absAng         = abs(intraPredAngle);
  Int signAng        = intraPredAngle < 0 ? -1 : 1;

  // Set bitshifts and scale the angle parameter to block size
  Int angTable[9]    = {0,    2,    5,   9,  13,  17,  21,  26,  32};
  Int invAngTable[9] = {0, 4096, 1638, 910, 630, 482, 390, 315, 256}; // (256 * 32) / Angle
  Int invAngle       = invAngTable[absAng];
  absAng             = angTable[absAng];
  intraPredAngle     = signAng * absAng;

  // Do the DC prediction
  if (modeDC)
  {
    Pel dcval = predIntraGetPredValDC(pSrc, srcStride, width, height, blkAboveAvailable, blkLeftAvailable);

    for (k=0;k<blkSize;k++)
    {
      for (l=0;l<blkSize;l++)
      {
        pDst[k*dstStride+l] = dcval;
      }
    }
  }

  // Do angular predictions
  else
  {
    Pel* refMain;
    Pel* refSide;
    Pel  refAbove[2*MAX_CU_SIZE+1];
    Pel  refLeft[2*MAX_CU_SIZE+1];

    // Initialise the Main and Left reference array.
    if (intraPredAngle < 0)
    {
      for (k=0;k<blkSize+1;k++)
      {
        refAbove[k+blkSize-1] = pSrc[k-srcStride-1];
      }
      for (k=0;k<blkSize+1;k++)
      {
        refLeft[k+blkSize-1] = pSrc[(k-1)*srcStride-1];
      }
      refMain = (modeVer ? refAbove : refLeft) + (blkSize-1);
      refSide = (modeVer ? refLeft : refAbove) + (blkSize-1);

      // Extend the Main reference to the left.
      Int invAngleSum    = 128;       // rounding for (shift by 8)
      for (k=-1; k>blkSize*intraPredAngle>>5; k--)
      {
        invAngleSum += invAngle;
        refMain[k] = refSide[invAngleSum>>8];
      }
    }
    else
    {
      for (k=0;k<2*blkSize+1;k++)
      {
        refAbove[k] = pSrc[k-srcStride-1];
      }
      for (k=0;k<2*blkSize+1;k++)
      {
        refLeft[k] = pSrc[(k-1)*srcStride-1];
      }
      refMain = modeVer ? refAbove : refLeft;
      refSide = modeVer ? refLeft  : refAbove;
    }

    if (intraPredAngle == 0)
    {
      for (k=0;k<blkSize;k++)
      {
        for (l=0;l<blkSize;l++)
        {
          pDst[k*dstStride+l] = refMain[l+1];
        }
      }

      if ( bFilter )
      {
        for (k=0;k<blkSize;k++)
        {
          pDst[k*dstStride] = Clip3(0, (1<<bitDepth)-1, pDst[k*dstStride] + (( refSide[k+1] - refSide[0] ) >> 1) );
        }
      }
    }
    else
    {
      Int deltaPos=0;
      Int deltaInt;
      Int deltaFract;
      Int refMainIndex;

      for (k=0;k<blkSize;k++)
      {
        deltaPos += intraPredAngle;
        deltaInt   = deltaPos >> 5;
        deltaFract = deltaPos & (32 - 1);

        if (deltaFract)
        {
          // Do linear filtering
          for (l=0;l<blkSize;l++)
          {
            refMainIndex        = l+deltaInt+1;
            pDst[k*dstStride+l] = (Pel) ( ((32-deltaFract)*refMain[refMainIndex]+deltaFract*refMain[refMainIndex+1]+16) >> 5 );
          }
        }
        else
        {
          // Just copy the integer samples
          for (l=0;l<blkSize;l++)
          {
            pDst[k*dstStride+l] = refMain[l+deltaInt+1];
          }
        }
      }
    }

    // Flip the block if this is the horizontal mode
    if (modeHor)
    {
      Pel  tmp;
      for (k=0;k<blkSize-1;k++)
      {
        for (l=k+1;l<blkSize;l++)
        {
          tmp                 = pDst[k*dstStride+l];
          pDst[k*dstStride+l] = pDst[l*dstStride+k];
          pDst[l*dstStride+k] = tmp;
        }
      }
    }
  }
}

Void TComPrediction::predIntraLumaAng(TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft )
{
  Pel *pDst = piPred;
  Int *ptrSrc;

  assert( g_aucConvertToBit[ iWidth ] >= 0 ); //   4x  4
  assert( g_aucConvertToBit[ iWidth ] <= 5 ); // 128x128
  assert( iWidth == iHeight  );

  ptrSrc = pcTComPattern->getPredictorPtr( uiDirMode, g_aucConvertToBit[ iWidth ] + 2, m_piYuvExt );

  // get starting pixel in block
  Int sw = 2 * iWidth + 1;

  // Create the prediction
  if ( uiDirMode == PLANAR_IDX )
  {
    xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight );
  }
  else
  {
    if ( (iWidth > 16) || (iHeight > 16) )
    {
      xPredIntraAng(g_bitDepthY, ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, false );
    }
    else
    {
      xPredIntraAng(g_bitDepthY, ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, true );

      if( (uiDirMode == DC_IDX ) && bAbove && bLeft )
      {
        xDCPredFiltering( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight);
      }
    }
  }
}

// Angular chroma
Void TComPrediction::predIntraChromaAng( Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft )
{
  Pel *pDst = piPred;
  Int *ptrSrc = piSrc;

  // get starting pixel in block
  Int sw = 2 * iWidth + 1;

  if ( uiDirMode == PLANAR_IDX )
  {
    xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight );
  }
  else
  {
    // Create the prediction
    xPredIntraAng(g_bitDepthC, ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, uiDirMode, bAbove, bLeft, false );
  }
}

/** Function for checking identical motion.
 * \param TComDataCU* pcCU
 * \param UInt PartAddr
 */
Bool TComPrediction::xCheckIdenticalMotion ( TComDataCU* pcCU, UInt PartAddr )
{
  if( pcCU->getSlice()->isInterB() && !pcCU->getSlice()->getPPS()->getWPBiPred() )
  {
    if( pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr) >= 0 && pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr) >= 0)
    {
      Int RefPOCL0 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_0, pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr))->getPOC();
      Int RefPOCL1 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_1, pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr))->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcCU->getCUMvField(REF_PIC_LIST_0)->getMv(PartAddr) == pcCU->getCUMvField(REF_PIC_LIST_1)->getMv(PartAddr))
      {
        return true;
      }
    }
  }
  return false;
}

#if INTRA_BL && !NO_RESIDUAL_FLAG_FOR_BLPRED
Void TComPrediction::getBaseBlk( TComDataCU* pcCU, TComYuv* pcYuvPred, Int iPartAddr, Int iWidth, Int iHeight )
{
  pcCU->getBaseLumaBlk( iWidth, iHeight, iPartAddr, pcYuvPred->getLumaAddr( iPartAddr ), pcYuvPred->getStride() );
  pcCU->getBaseChromaBlk( iWidth >> 1, iHeight >> 1, iPartAddr, pcYuvPred->getCbAddr( iPartAddr ), pcYuvPred->getCStride(), 0 );
  pcCU->getBaseChromaBlk( iWidth >> 1, iHeight >> 1, iPartAddr, pcYuvPred->getCrAddr( iPartAddr ), pcYuvPred->getCStride(), 1 );
}
#endif

Void TComPrediction::motionCompensation ( TComDataCU* pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList, Int iPartIdx )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  if ( iPartIdx >= 0 )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );
    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
    return;
  }

  for ( iPartIdx = 0; iPartIdx < pcCU->getNumPartInter(); iPartIdx++ )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );

    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
  }
  return;
}
#if DEC_EILR_BLK
Void TComPrediction::motionCompensationEL ( TComDataCU* pcCU, TComYuv* pcYuvPred, TComPic *pEILRPic
#if ILR_TM_ENH_UCM
                                            ,TComPicSym* pStoredPicSym
#endif
  )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  for ( Int iPartIdx = 0; iPartIdx < pcCU->getNumPartInter(); iPartIdx++ )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );

    if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
    {
      xPredInterUniEL (pEILRPic, pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred
#if ILR_TM_ENH_UCM
        ,false, pStoredPicSym
#endif
        );
    }
    else
    {
      xPredInterBiEL  (pEILRPic, pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred
#if ILR_TM_ENH_UCM
        ,pStoredPicSym
#endif
        );
    }
  }
  return;
}
#endif

Void TComPrediction::xPredInterUni ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi )
{
  Int         iRefIdx     = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );           assert (iRefIdx >= 0);
  TComMv      cMv         = pcCU->getCUMvField( eRefPicList )->getMv( uiPartAddr );
  pcCU->clipMv(cMv);

#if REF_IDX_ME_ZEROMV
  assert( ( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) && cMv.getHor() == 0 && cMv.getVer() == 0 ) || pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) == false );
#endif

  xPredInterLumaBlk  ( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
  xPredInterChromaBlk( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
}

#if DEC_EILR_BLK
Void TComPrediction::xPredInterUniEL ( TComPic *pcEILRPic, TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi
#if ILR_TM_ENH_UCM
                                       ,TComPicSym* pStoredPicSym
#endif
  )
{
  Int         iRefIdx     = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );           assert (iRefIdx >= 0);
  TComMv      cMv         = pcCU->getCUMvField( eRefPicList )->getMv( uiPartAddr );
  pcCU->clipMv(cMv);

#if REF_IDX_ME_ZEROMV
  assert( ( pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) && cMv.getHor() == 0 && cMv.getVer() == 0 ) || pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) == false );
#endif

  if(pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->isILR(pcCU->getLayerId()) && pcCU->getSlice()->getRefPic(eRefPicList, iRefIdx)->getEnhILRType())
  {
    xPredInterBlkILR  ( pcEILRPic, pcCU, uiPartAddr, iWidth, iHeight, rpcYuvPred, bi
#if ILR_TM_ENH_UCM
                        ,pStoredPicSym
#endif
      );
  }
  else
  {
   xPredInterLumaBlk  ( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
   xPredInterChromaBlk( pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
  }
}
#endif
Void TComPrediction::xPredInterBi ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred )
{
  TComYuv* pcMbYuv;
  Int      iRefIdx[2] = {-1, -1};

  for ( Int iRefList = 0; iRefList < 2; iRefList++ )
  {
    RefPicList eRefPicList = (iRefList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
    iRefIdx[iRefList] = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );

    if ( iRefIdx[iRefList] < 0 )
    {
      continue;
    }

    assert( iRefIdx[iRefList] < pcCU->getSlice()->getNumRefIdx(eRefPicList) );

    pcMbYuv = &m_acYuvPred[iRefList];
    if( pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr ) >= 0 && pcCU->getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiPartAddr ) >= 0 )
    {
      xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
    }
    else
    {
      if ( ( pcCU->getSlice()->getPPS()->getUseWP()       && pcCU->getSlice()->getSliceType() == P_SLICE ) || 
           ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE ) )
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
      }
      else
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv );
      }
    }
  }

  if ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE  )
  {
    xWeightedPredictionBi( pcCU, &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }  
  else if ( pcCU->getSlice()->getPPS()->getUseWP() && pcCU->getSlice()->getSliceType() == P_SLICE )
  {
    xWeightedPredictionUni( pcCU, &m_acYuvPred[0], uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, rpcYuvPred ); 
  }
  else
  {
    xWeightedAverage( &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
}
#if DEC_EILR_BLK
Void TComPrediction::xPredInterBiEL ( TComPic *pEILRPic, TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred
#if ILR_TM_ENH_UCM
                                     ,TComPicSym* pStoredPicSym
#endif
  )
{
  TComYuv* pcMbYuv;
  Int      iRefIdx[2] = {-1, -1};

  for ( Int iRefList = 0; iRefList < 2; iRefList++ )
  {
    RefPicList eRefPicList = (iRefList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
    iRefIdx[iRefList] = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );

    if ( iRefIdx[iRefList] < 0 )
    {
      continue;
    }

    assert( iRefIdx[iRefList] < pcCU->getSlice()->getNumRefIdx(eRefPicList) );

    pcMbYuv = &m_acYuvPred[iRefList];
    if( pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr ) >= 0 && pcCU->getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiPartAddr ) >= 0 )
    {
      xPredInterUniEL ( pEILRPic, pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true
#if ILR_TM_ENH_UCM
                        ,pStoredPicSym
#endif
                      );
    }
    else
    {
      if ( ( pcCU->getSlice()->getPPS()->getUseWP()       && pcCU->getSlice()->getSliceType() == P_SLICE ) || 
           ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE ) )
      {
        xPredInterUniEL ( pEILRPic, pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true
#if ILR_TM_ENH_UCM
                        ,pStoredPicSym
#endif
                        );
      }
      else
      {
        xPredInterUniEL ( pEILRPic, pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, false
#if ILR_TM_ENH_UCM
                        ,pStoredPicSym
#endif
                        );
      }
    }
  }

  if ( pcCU->getSlice()->getPPS()->getWPBiPred() && pcCU->getSlice()->getSliceType() == B_SLICE  )
  {
    xWeightedPredictionBi( pcCU, &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }  
  else if ( pcCU->getSlice()->getPPS()->getUseWP() && pcCU->getSlice()->getSliceType() == P_SLICE )
  {
    xWeightedPredictionUni( pcCU, &m_acYuvPred[0], uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, rpcYuvPred ); 
  }
  else
  {
    xWeightedAverage( &m_acYuvPred[0], &m_acYuvPred[1], iRefIdx[0], iRefIdx[1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
}
#endif
/**
 * \brief Generate motion-compensated luma block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */
Void TComPrediction::xPredInterLumaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int refStride = refPic->getStride();  
  Int refOffset = ( mv->getHor() >> 2 ) + ( mv->getVer() >> 2 ) * refStride;
  Pel *ref      = refPic->getLumaAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Int dstStride = dstPic->getStride();
  Pel *dst      = dstPic->getLumaAddr( partAddr );
  
  Int xFrac = mv->getHor() & 0x3;
  Int yFrac = mv->getVer() & 0x3;

  if ( yFrac == 0 )
  {
    m_if.filterHorLuma( ref, refStride, dst, dstStride, width, height, xFrac,       !bi );
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerLuma( ref, refStride, dst, dstStride, width, height, yFrac, true, !bi );
  }
  else
  {
    Int tmpStride = m_filteredBlockTmp[0].getStride();
    Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

    Int filterSize = NTAPS_LUMA;
    Int halfFilterSize = ( filterSize >> 1 );

    m_if.filterHorLuma(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, xFrac, false     );
    m_if.filterVerLuma(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              yFrac, false, !bi);    
  }
}
#if ILR_TM_ENH
Void TComPrediction::xPredInterLumaBlkByPosBilinear(TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int refStride = refPic->getStride();  
  TComMv  mvImp = *mv;

  Int refOffset = ( mvImp.getHor() >> 2 ) + ( mvImp.getVer() >> 2 ) * refStride;
  Pel *ref      = refPic->getLumaAddr() + iBlkY*refStride + iBlkX + refOffset;
  
  Int dstStride = dstPic->getStride();
  Pel *dst      = dstPic->getLumaAddr( 0 );
  
  Int xFrac = mvImp.getHor() & 0x3;
  Int yFrac = mvImp.getVer() & 0x3;

  if ( yFrac == 0 )
  {
    m_if.filterHorBilinear( ref, refStride, dst, dstStride, width, height, (xFrac<<1),       !bi );
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerBilinear( ref, refStride, dst, dstStride, width, height, (yFrac<<1), true, !bi );
  }
  else
  {
    Int tmpStride = m_filteredBlockTmp[0].getStride();
    Short *tmp    = m_filteredBlockTmp[0].getLumaAddr();

    Int filterSize = NTAPS_BILINEAR;
    Int halfFilterSize = ( filterSize >> 1 );

    m_if.filterHorBilinear(ref - (halfFilterSize-1)*refStride, refStride, tmp, tmpStride, width, height+filterSize-1, (xFrac<<1), false     );
    m_if.filterVerBilinear(tmp + (halfFilterSize-1)*tmpStride, tmpStride, dst, dstStride, width, height,              (yFrac<<1), false, !bi);    
  }
}
#endif

/**
 * \brief Generate motion-compensated chroma block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */
Void TComPrediction::xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getCStride();
  Int     dstStride  = dstPic->getCStride();
  
  Int     refOffset  = (mv->getHor() >> 3) + (mv->getVer() >> 3) * refStride;
  
  Pel*    refCb     = refPic->getCbAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  Pel*    refCr     = refPic->getCrAddr( cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Pel* dstCb = dstPic->getCbAddr( partAddr );
  Pel* dstCr = dstPic->getCrAddr( partAddr );
  
  Int     xFrac  = mv->getHor() & 0x7;
  Int     yFrac  = mv->getVer() & 0x7;
  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;
  
  Int     extStride = m_filteredBlockTmp[0].getStride();
  Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();
  
  Int filterSize = NTAPS_CHROMA;
  
  Int halfFilterSize = (filterSize>>1);
  
  if ( yFrac == 0 )
  {
    m_if.filterHorChroma(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
    m_if.filterHorChroma(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerChroma(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
    m_if.filterVerChroma(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
  }
  else
  {
    m_if.filterHorChroma(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);
    
    m_if.filterHorChroma(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerChroma(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
  }
}
#if ILR_TM_ENH
Void TComPrediction::xPredInterChromaBlkByPosBilinear(TComPicYuv *refPic, Int iBlkX, Int iBlkY, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getCStride();
  Int     dstStride  = dstPic->getCStride();
  TComMv  mvImp = *mv;
  
  Int     refOffset  = (mvImp.getHor() >> 3) + (mvImp.getVer() >> 3) * refStride;
  
  Pel*    refCb     = refPic->getCbAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
  Pel*    refCr     = refPic->getCrAddr() + (iBlkY>>1)*refStride+(iBlkX>>1) + refOffset;
  
  Pel* dstCb = dstPic->getCbAddr( 0 );
  Pel* dstCr = dstPic->getCrAddr( 0 );
  
  Int     xFrac  = mvImp.getHor() & 0x7;
  Int     yFrac  = mvImp.getVer() & 0x7;
  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;
  
  Int     extStride = m_filteredBlockTmp[0].getStride();
  Short*  extY      = m_filteredBlockTmp[0].getLumaAddr();
  
  Int filterSize = NTAPS_BILINEAR;
  
  Int halfFilterSize = (filterSize>>1);
  
  if ( yFrac == 0 )
  {
    m_if.filterHorBilinear(refCb, refStride, dstCb,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
    m_if.filterHorBilinear(refCr, refStride, dstCr,  dstStride, cxWidth, cxHeight, xFrac, !bi);    
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVerBilinear(refCb, refStride, dstCb, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
    m_if.filterVerBilinear(refCr, refStride, dstCr, dstStride, cxWidth, cxHeight, yFrac, true, !bi);    
  }
  else
  {
    m_if.filterHorBilinear(refCb - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerBilinear(extY  + (halfFilterSize-1)*extStride, extStride, dstCb, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);
    
    m_if.filterHorBilinear(refCr - (halfFilterSize-1)*refStride, refStride, extY,  extStride, cxWidth, cxHeight+filterSize-1, xFrac, false);
    m_if.filterVerBilinear(extY  + (halfFilterSize-1)*extStride, extStride, dstCr, dstStride, cxWidth, cxHeight  , yFrac, false, !bi);    
  }
}
#endif

#if DEC_EILR_BLK
Void TComPrediction::xPredInterBlkILR  ( TComPic *pcEILRPic, TComDataCU* pcCU, UInt uiPartAddr, Int iBWidth, Int iBHeight, TComYuv*& rpcYuvPred, Bool bi
#if ILR_TM_ENH_UCM
                                        ,TComPicSym* pStoredPicSym
#endif
                                        )
{
  Int iSX = pcCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiPartAddr] ]; //the start position of MC block;
  Int iSY = pcCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiPartAddr] ];
  Int iEX = iSX+iBWidth;  //end position of MC block;
  Int iEY = iSY+iBHeight;
  //TComSlice *pcSlice = pcCU->getSlice();
  TComPic *pcPicSrc = pcCU->getPic();

#if ILR_TM_ENH_UCM
#if ILR_TM_ENH_UCM==3
  Int  iPUWidth = 16;
  Int  iPUHeight = 16;
#else
  Int iMinPUWidth  = g_uiMaxCUWidth>>g_uiMaxCUDepth;
	Int iMinPUHeight = g_uiMaxCUHeight>>g_uiMaxCUDepth;
  Int  iPUWidth = iMinPUWidth*m_iSRNum[0];
  Int  iPUHeight = iMinPUHeight*m_iSRNum[1];
#endif
#else
  Int  iPUWidth = (m_iSRNum[0]<<3)&(~0xf); //16; 
  Int  iPUHeight = (m_iSRNum[1]<<3)&(~0xf); //16; 
#endif
  TComMv cBlkPos, cBlkSize, cStartPos;
  Int iPBY = iSY/iPUHeight*iPUHeight;
  Int iPBX = iSX/iPUWidth*iPUWidth;
  Int iBlkHeight;
  cStartPos.set(iSX, iSY);
  for ( Int iBlkY =iPBY; iBlkY < iEY; iBlkY += iBlkHeight)
  {
    iBlkHeight = min((Int)(g_uiMaxCUHeight - (iBlkY%g_uiMaxCUHeight)), (iEY - iBlkY));
    Int iBlkWidth;
    for(Int iBlkX =iPBX; iBlkX < iEX; iBlkX += iBlkWidth)
    {
      iBlkWidth = min((Int)(g_uiMaxCUWidth-(iBlkX%g_uiMaxCUWidth)), (Int)(iEX - iBlkX));

      //one LCU;
      Int iSubBlkHeight;
      for(Int iSubBlkY=iBlkY; iSubBlkY<iBlkY+iBlkHeight; iSubBlkY += iSubBlkHeight)
      {
        iSubBlkHeight = min(iPUHeight-(iSubBlkY%iPUHeight), (iBlkY+iBlkHeight - iSubBlkY));        
        if(iSubBlkY >= iEY || iSubBlkY+iSubBlkHeight<=iSY)
          continue;
        Int iPosY = max(iSY, iSubBlkY);
        Int iSizeY = min(iEY, iSubBlkY+iSubBlkHeight)-iPosY;
        

        Int iSubBlkWidth;
        for(Int iSubBlkX =iBlkX; iSubBlkX < iBlkX+iBlkWidth; iSubBlkX += iSubBlkWidth)
        {
          iSubBlkWidth = min(iPUWidth-(iSubBlkX%iPUWidth), (iBlkX+iBlkWidth - iSubBlkX)); 
          if(iSubBlkX >= iEX || iSubBlkX+iSubBlkWidth<=iSX) 
            continue;
          Int iPosX = max(iSubBlkX, iSX);
          
          cBlkSize.set(min(iEX, iSubBlkX+iSubBlkWidth)-iPosX, iSizeY);
          cBlkPos.set(iPosX, iPosY);

#if ILR_TM_ENH_WEIGHT_SIGNAL
          getEnhBlockMPUWeightDec(pcPicSrc, pcEILRPic, rpcYuvPred, uiPartAddr, &cStartPos, &cBlkPos, &cBlkSize
#if ILR_TM_ENH_UCM
            , pStoredPicSym
#endif
            ); 
#else
#error "has to be implemented"
#endif
        }
      }//end LCU;
    }
  }

  if(bi)
  {
    Int shift = IF_INTERNAL_PREC - g_bitDepthY;
    Pel *src = rpcYuvPred->getLumaAddr(uiPartAddr);
    Int srcStride = rpcYuvPred->getStride(); 
    for (Int row = 0; row < iBHeight; row++)
    {
      for (Int col = 0; col < iBWidth; col++)
      {
        src[col] = (src[col] << shift) - (Short)IF_INTERNAL_OFFS;
      }      
      src += srcStride;
    }
    //shift = IF_INTERNAL_PREC - g_bitDepthC;
    Pel *src2 = rpcYuvPred->getCrAddr(uiPartAddr);
    src = rpcYuvPred->getCbAddr(uiPartAddr);    
    srcStride = rpcYuvPred->getCStride(); 
    for (Int row = 0; row < (iBHeight>>1); row++)
    {
      for (Int col = 0; col < (iBWidth>>1); col++)
      {
        src[col] = (src[col] << shift) - (Short)IF_INTERNAL_OFFS;
        src2[col] = (src2[col] << shift) - (Short)IF_INTERNAL_OFFS;
      }      
      src += srcStride;
      src2 += srcStride;
    }          
  }
}
#endif

Void TComPrediction::xWeightedAverage( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartIdx, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst )
{
  if( iRefIdx0 >= 0 && iRefIdx1 >= 0 )
  {
    rpcYuvDst->addAvg( pcYuvSrc0, pcYuvSrc1, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 >= 0 && iRefIdx1 <  0 )
  {
    pcYuvSrc0->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 <  0 && iRefIdx1 >= 0 )
  {
    pcYuvSrc1->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
}

// AMVP
Void TComPrediction::getMvPredAMVP( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, TComMv& rcMvPred )
{
  AMVPInfo* pcAMVPInfo = pcCU->getCUMvField(eRefPicList)->getAMVPInfo();
  if( pcAMVPInfo->iN <= 1 )
  {
    rcMvPred = pcAMVPInfo->m_acMvCand[0];

    pcCU->setMVPIdxSubParts( 0, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    pcCU->setMVPNumSubParts( pcAMVPInfo->iN, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    return;
  }

  assert(pcCU->getMVPIdx(eRefPicList,uiPartAddr) >= 0);
  rcMvPred = pcAMVPInfo->m_acMvCand[pcCU->getMVPIdx(eRefPicList,uiPartAddr)];
  return;
}

/** Function for deriving planar intra prediction.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 *
 * This function derives the prediction samples for planar mode (intra coding).
 */
Void TComPrediction::xPredIntraPlanar( Int* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height )
{
  assert(width == height);

  Int k, l, bottomLeft, topRight;
  Int horPred;
  Int leftColumn[MAX_CU_SIZE], topRow[MAX_CU_SIZE], bottomRow[MAX_CU_SIZE], rightColumn[MAX_CU_SIZE];
  UInt blkSize = width;
  UInt offset2D = width;
  UInt shift1D = g_aucConvertToBit[ width ] + 2;
  UInt shift2D = shift1D + 1;

  // Get left and above reference column and row
  for(k=0;k<blkSize+1;k++)
  {
    topRow[k] = pSrc[k-srcStride];
    leftColumn[k] = pSrc[k*srcStride-1];
  }

  // Prepare intermediate variables used in interpolation
  bottomLeft = leftColumn[blkSize];
  topRight   = topRow[blkSize];
  for (k=0;k<blkSize;k++)
  {
    bottomRow[k]   = bottomLeft - topRow[k];
    rightColumn[k] = topRight   - leftColumn[k];
    topRow[k]      <<= shift1D;
    leftColumn[k]  <<= shift1D;
  }

  // Generate prediction signal
  for (k=0;k<blkSize;k++)
  {
    horPred = leftColumn[k] + offset2D;
    for (l=0;l<blkSize;l++)
    {
      horPred += rightColumn[k];
      topRow[l] += bottomRow[l];
      rpDst[k*dstStride+l] = ( (horPred + topRow[l]) >> shift2D );
    }
  }
}

/** Function for filtering intra DC predictor.
 * \param pSrc pointer to reconstructed sample array
 * \param iSrcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param iDstStride the stride of the prediction sample array
 * \param iWidth the width of the block
 * \param iHeight the height of the block
 *
 * This function performs filtering left and top edges of the prediction samples for DC mode (intra coding).
 */
Void TComPrediction::xDCPredFiltering( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight )
{
  Pel* pDst = rpDst;
  Int x, y, iDstStride2, iSrcStride2;

  // boundary pixels processing
  pDst[0] = (Pel)((pSrc[-iSrcStride] + pSrc[-1] + 2 * pDst[0] + 2) >> 2);

  for ( x = 1; x < iWidth; x++ )
  {
    pDst[x] = (Pel)((pSrc[x - iSrcStride] +  3 * pDst[x] + 2) >> 2);
  }

  for ( y = 1, iDstStride2 = iDstStride, iSrcStride2 = iSrcStride-1; y < iHeight; y++, iDstStride2+=iDstStride, iSrcStride2+=iSrcStride )
  {
    pDst[iDstStride2] = (Pel)((pSrc[iSrcStride2] + 3 * pDst[iDstStride2] + 2) >> 2);
  }

  return;
}

#if SVC_UPSAMPLING
#if SCALED_REF_LAYER_OFFSETS
Void TComPrediction::upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic, const Window window)
{
  m_cUsf.upsampleBasePic( pcUsPic, pcBasePic, pcTempPic, window);
}
#else
Void TComPrediction::upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic)
{
  m_cUsf.upsampleBasePic( pcUsPic, pcBasePic, pcTempPic);
}
#endif
#endif

#if ILR_TM_ENH
static Void copyBlk(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight)
{
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));

  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iBlkX >>= 1;
  iBlkY >>= 1;
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  
  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
  pSrc = pcPicYuvSrc->getCrAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
}
/*
Void TComPrediction::calcEnhBlk(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight) //, UInt uiAbsPartIdx)
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
  Int iStrideMCELx = m_pEnhPred[1]->getStride();

  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
  Pel *pMCELx = m_pEnhPred[1]->getLumaAddr();

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i+=2)
    {
      pDst[i] = ClipY(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
      pDst[i+1] = ClipY(pSrc[i+1]+pMCELx[i+1]-pMCUpBLx[i+1]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCELx += iStrideMCELx;
    pMCUpBLx += iStrideMCUpBLx;
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
  iStrideMCELx = m_pEnhPred[1]->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
  pMCELx = m_pEnhPred[1]->getCbAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
    }

    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCELx += iStrideMCELx;
    pMCUpBLx += iStrideMCUpBLx;
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
  pMCELx = m_pEnhPred[1]->getCrAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCELx += iStrideMCELx;
    pMCUpBLx += iStrideMCUpBLx;
  }
}
*/
#if DEC_EILR_BLK
static Void copyBlkDec(TComPicYuv *pcPicYuvSrc, TComYuv *&pcPicYuvDst, UInt uiPartAddr, TComMv *pSPos, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight)
{
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr(uiPartAddr) + (iBlkY-pSPos->getVer())*iStrideDst + (iBlkX-pSPos->getHor());
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));

  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iBlkX >>= 1;
  iBlkY >>= 1;
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  
  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr(uiPartAddr) + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
  pSrc = pcPicYuvSrc->getCrAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr(uiPartAddr)  + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
}
#if (ILR_TM_ENH_EXPLICIT_MODE==1)
static Void copyBlkOneCompDec(TComPicYuv *pcPicYuvSrc, TComYuv *&pcPicYuvDst, UInt uiPartAddr, TComMv *pSPos, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx)
{
  Int iStrideSrc;
  Int iStrideDst;
  Pel *pSrc;
  Pel *pDst;

  if(!iCompIdx)
  {
    iStrideSrc = pcPicYuvSrc->getStride();
    iStrideDst = pcPicYuvDst->getStride();
    pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getLumaAddr(uiPartAddr) + (iBlkY-pSPos->getVer())*iStrideDst + (iBlkX-pSPos->getHor());
  }
  else if(iCompIdx==1)
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iBlkX >>= 1;
    iBlkY >>= 1;
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCbAddr(uiPartAddr) + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
  }
  else
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iBlkX >>= 1;
    iBlkY >>= 1;
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    pSrc = pcPicYuvSrc->getCrAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCrAddr(uiPartAddr) + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
  }

  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
}
#endif
Void TComPrediction::calcEnhBlkOneCompDec(TComPicYuv *pcPicYuvSrc, TComYuv *&pcPicYuvDst, UInt uiPartAddr, TComMv *pSPos, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx) 
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iStrideSrc;
  Int iStrideDst;
  Int iStrideMCUpBLx;
  Int iStrideMCELx;
  Pel *pSrc;
  Pel *pDst;
  Pel *pMCUpBLx;
  Pel *pMCELx;

  if(iCompIdx==0)
  {
    iStrideSrc = pcPicYuvSrc->getStride();
    iStrideDst = pcPicYuvDst->getStride();
    iStrideMCUpBLx = m_pEnhPred[0]->getStride();
    iStrideMCELx = m_pEnhPred[1]->getStride();
    pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getLumaAddr(uiPartAddr) + (iBlkY-pSPos->getVer())*iStrideDst + (iBlkX-pSPos->getHor());
    pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
    pMCELx = m_pEnhPred[1]->getLumaAddr();
  }
  else if(iCompIdx==1)
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
    iStrideMCELx = m_pEnhPred[1]->getCStride();
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    iBlkX >>= 1;
    iBlkY >>= 1;

    pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCbAddr(uiPartAddr) + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
    pMCUpBLx = m_pEnhPred[0]->getCbAddr();
    pMCELx = m_pEnhPred[1]->getCbAddr();
  }
  else
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
    iStrideMCELx = m_pEnhPred[1]->getCStride();
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    iBlkX >>= 1;
    iBlkY >>= 1;

    pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCrAddr(uiPartAddr) + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
    pMCUpBLx = m_pEnhPred[0]->getCrAddr();
    pMCELx = m_pEnhPred[1]->getCrAddr();
  }

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCELx += iStrideMCELx;
    pMCUpBLx += iStrideMCUpBLx;
  }
}
#endif
#if ILR_TM_ENH_EXPLICIT_MODE
Void TComPrediction::allocateEILRBlockMode(UInt uiSWidth, UInt uiSHeight, UInt uiMinPUWidth, UInt uiMinPUHeight)
{

  if(!m_pchEILRBlockMode)
  { 
    UInt uiWidthPU = (uiSWidth+uiMinPUWidth-1)/uiMinPUWidth;
    UInt uiHeightPU = (uiSHeight+uiMinPUHeight-1)/uiMinPUHeight;
    m_pchEILRBlockMode = (Char **)malloc(uiHeightPU * sizeof(Char*));
    m_pchEILRBlockMode[0] = (Char *)xMalloc(Char, uiWidthPU*uiHeightPU);
   
    for(Int i=1; i<uiHeightPU; i++)
     m_pchEILRBlockMode[i] = m_pchEILRBlockMode[i-1] + uiWidthPU;
  }

}

Void TComPrediction::copyBlkOneComp(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx)
{
  Int iStrideSrc;
  Int iStrideDst;
  Pel *pSrc;
  Pel *pDst;

  if(!iCompIdx)
  {
    iStrideSrc = pcPicYuvSrc->getStride();
    iStrideDst = pcPicYuvDst->getStride();
    pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  }
  else if(iCompIdx==1)
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iBlkX >>= 1;
    iBlkY >>= 1;
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  }
  else
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iBlkX >>= 1;
    iBlkY >>= 1;
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    pSrc = pcPicYuvSrc->getCrAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCrAddr() + iBlkY*iStrideDst + iBlkX;
  }

  for(Int j=0; j<iBlkHeight; j++)
    memcpy(pDst+j*iStrideDst, pSrc+j*iStrideSrc, iBlkWidth*sizeof(Pel));
}

Void TComPrediction::calcEnhBlkOneComp(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx) 
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iStrideSrc;
  Int iStrideDst;
  Int iStrideMCUpBLx;
  Int iStrideMCELx;
  Pel *pSrc;
  Pel *pDst;
  Pel *pMCUpBLx;
  Pel *pMCELx;

  if(iCompIdx==0)
  {
    iStrideSrc = pcPicYuvSrc->getStride();
    iStrideDst = pcPicYuvDst->getStride();
    iStrideMCUpBLx = m_pEnhPred[0]->getStride();
    iStrideMCELx = m_pEnhPred[1]->getStride();
    pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
    pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
    pMCELx = m_pEnhPred[1]->getLumaAddr();
  }
  else if(iCompIdx==1)
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
    iStrideMCELx = m_pEnhPred[1]->getCStride();
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    iBlkX >>= 1;
    iBlkY >>= 1;

    pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
    pMCUpBLx = m_pEnhPred[0]->getCbAddr();
    pMCELx = m_pEnhPred[1]->getCbAddr();
  }
  else
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
    iStrideMCELx = m_pEnhPred[1]->getCStride();
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    iBlkX >>= 1;
    iBlkY >>= 1;

    pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
    pMCUpBLx = m_pEnhPred[0]->getCrAddr();
    pMCELx = m_pEnhPred[1]->getCrAddr();
  }

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+pMCELx[i]-pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCELx += iStrideMCELx;
    pMCUpBLx += iStrideMCUpBLx;
  }
}

Void TComPrediction::calcEnhBlk2OneComp(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iCompIdx)
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iStrideSrc;
  Int iStrideDst;
  Pel *pSrc;
  Pel *pDst;

  if(iCompIdx == 0)
  {
    iStrideSrc = pcPicYuvSrc->getStride();
    iStrideDst = pcPicYuvDst->getStride();
    pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  }
  else if(iCompIdx ==1)
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    iBlkX >>= 1;
    iBlkY >>= 1;

    pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  }
  else
  {
    iStrideSrc = pcPicYuvSrc->getCStride();
    iStrideDst = pcPicYuvDst->getCStride();
    iBlkWidth >>=1;
    iBlkHeight >>=1;
    iBlkX >>= 1;
    iBlkY >>= 1;

    pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
    pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  }

  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+pDst[i]);
      //pDst[i+1] = Clip(pSrc[i+1]+pDst[i+1]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
  }
}
#endif  //ILR_TM_ENH_EXPLICIT_MODE

#if ILR_TM_ENH_WEIGHT_SIGNAL
Void TComPrediction::calcEnhBlkModeDiff(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
                                        ,Int iOffset
#endif
                                       )
{
  //dst = up(BL) + (1-weight)*(MC(ELx)-MC(Up(BLx)));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
  Pel *pMCELx = m_pEnhPred[1]->getLumaAddr();
  Int iStrideMCELx = m_pEnhPred[1]->getStride();

  if(iWeight)
  {
    iWeight = WEIGHT_SCALE - iWeight;
    if(iWeight)
    {
      for(Int j=0; j<iBlkHeight; j++)
      {
        for(Int i=0; i<iBlkWidth; i+=2)
        {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
          pDst[i] = ClipY(pSrc[i]+ ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET + iOffset)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ ((iWeight*(pMCELx[i+1] - pMCUpBLx[i+1])+WEIGHT_SCALE_OFFSET + iOffset)>>WEIGHT_SCALE_SHIFT));
#else
          pDst[i] = ClipY(pSrc[i]+ ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ ((iWeight*(pMCELx[i+1] - pMCUpBLx[i+1])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
#endif
        }
        pDst += iStrideDst;
        pSrc += iStrideSrc;
        pMCUpBLx += iStrideMCUpBLx;
        pMCELx += iStrideMCELx;
      }
    }
    else
    { //copy;
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
      iOffset = (iOffset + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT;
#endif
      for(Int j=0; j<iBlkHeight; j++)
      {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        for(Int i=0; i<iBlkWidth; i+=2)
        {
          pDst[i] = ClipY(pSrc[i]+ iOffset);
          pDst[i+1] = ClipY(pSrc[i+1]+ iOffset);
        }
#else
        memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel)); 
#endif
        pDst += iStrideDst;
        pSrc += iStrideSrc;
      }
    }
  }
  else
  {
    //no weight;
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
      iOffset = (iOffset + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT;
#endif
      for(Int j=0; j<iBlkHeight; j++)
      {
        for(Int i=0; i<iBlkWidth; i+=2)
        {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
          pDst[i] = ClipY(pSrc[i]+ (pMCELx[i] - pMCUpBLx[i]) + iOffset);
          pDst[i+1] = ClipY(pSrc[i+1]+ (pMCELx[i+1] - pMCUpBLx[i+1]) + iOffset);
#else
          pDst[i] = ClipY(pSrc[i]+ (pMCELx[i] - pMCUpBLx[i]));
          pDst[i+1] = ClipY(pSrc[i+1]+ (pMCELx[i+1] - pMCUpBLx[i+1]));
#endif
        }
        pDst += iStrideDst;
        pSrc += iStrideSrc;
        pMCUpBLx += iStrideMCUpBLx;
        pMCELx += iStrideMCELx;
      }
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
  iStrideMCELx = m_pEnhPred[1]->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
  pMCELx = m_pEnhPred[1]->getCbAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ClipC(pSrc[i] + ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = ClipC(pSrc[i] + pMCELx[i] - pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
    pMCELx += iStrideMCELx;
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  pMCELx = m_pEnhPred[1]->getCrAddr();
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ClipC(pSrc[i] + ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = ClipC(pSrc[i] + pMCELx[i] - pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
    pMCELx += iStrideMCELx;
  }
}

#if DEC_EILR_BLK
Void TComPrediction::calcEnhBlkModeDiffDec(TComPicYuv *pcPicYuvSrc, TComYuv *&pcPicYuvDst, UInt uiPartAddr, TComMv *pSPos, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
                                           , Int iOffset
#endif
                                          )
{
  //dst = up(BL) + (1-weight)*(MC(ELx)-MC(Up(BLx)));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Int iStrideMCUpBLx = m_pEnhPred[0]->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr(uiPartAddr) + (iBlkY - pSPos->getVer())*iStrideDst + (iBlkX - pSPos->getHor());
  Pel *pMCUpBLx = m_pEnhPred[0]->getLumaAddr();
  Pel *pMCELx = m_pEnhPred[1]->getLumaAddr();
  Int iStrideMCELx = m_pEnhPred[1]->getStride();
  if(iWeight)
  {
    iWeight = WEIGHT_SCALE - iWeight;
    if(iWeight==1)
    {
      for(Int j=0; j<iBlkHeight; j++)
      {
        for(Int i=0; i<iBlkWidth; i+=2)
        {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
          pDst[i] = ClipY(pSrc[i]+ (((pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET + iOffset)>>WEIGHT_SCALE_SHIFT) );
          pDst[i+1] = ClipY(pSrc[i+1]+ (((pMCELx[i+1] - pMCUpBLx[i+1])+WEIGHT_SCALE_OFFSET + iOffset)>>WEIGHT_SCALE_SHIFT));
#else
          pDst[i] = ClipY(pSrc[i]+ (((pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ (((pMCELx[i+1] - pMCUpBLx[i+1])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
#endif
        }
        pDst += iStrideDst;
        pSrc += iStrideSrc;
        pMCUpBLx += iStrideMCUpBLx;
        pMCELx += iStrideMCELx;
      }
    }
    else if(iWeight)
    {
      for(Int j=0; j<iBlkHeight; j++)
      {
        for(Int i=0; i<iBlkWidth; i+=2)
        {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
          pDst[i] = ClipY(pSrc[i]+ ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET+iOffset)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ ((iWeight*(pMCELx[i+1] - pMCUpBLx[i+1])+WEIGHT_SCALE_OFFSET+iOffset)>>WEIGHT_SCALE_SHIFT));
#else
          pDst[i] = ClipY(pSrc[i]+ ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ ((iWeight*(pMCELx[i+1] - pMCUpBLx[i+1])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
#endif
        }
        pDst += iStrideDst;
        pSrc += iStrideSrc;
        pMCUpBLx += iStrideMCUpBLx;
        pMCELx += iStrideMCELx;
      }
    }
    else
    {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
      iOffset = (iOffset + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT;
#endif
      for(Int j=0; j<iBlkHeight; j++)
      {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        for(Int i=0; i<iBlkWidth; i+=2)
        {
          pDst[i] = ClipY(pSrc[i]+ iOffset);
          pDst[i+1] = ClipY(pSrc[i+1]+ iOffset);
        }
#else
        memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
        pDst += iStrideDst;
        pSrc += iStrideSrc;
      }
    }
  }
  else
  {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
    iOffset = (iOffset + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT;
#endif
    for(Int j=0; j<iBlkHeight; j++)
    {
      for(Int i=0; i<iBlkWidth; i+=2)
      {
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        pDst[i] = ClipY(pSrc[i]+ (pMCELx[i] - pMCUpBLx[i])+iOffset);
        pDst[i+1] = ClipY(pSrc[i+1]+ (pMCELx[i+1] - pMCUpBLx[i+1])+iOffset);
#else
        pDst[i] = ClipY(pSrc[i]+ (pMCELx[i] - pMCUpBLx[i]));
        pDst[i+1] = ClipY(pSrc[i+1]+ (pMCELx[i+1] - pMCUpBLx[i+1]));
#endif
      }
      pDst += iStrideDst;
      pSrc += iStrideSrc;
      pMCUpBLx += iStrideMCUpBLx;
      pMCELx += iStrideMCELx;
    }
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iStrideMCUpBLx = m_pEnhPred[0]->getCStride();
  iStrideMCELx = m_pEnhPred[1]->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr(uiPartAddr) + (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
  pMCUpBLx = m_pEnhPred[0]->getCbAddr();
  pMCELx = m_pEnhPred[1]->getCbAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ClipC(pSrc[i] + ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = ClipC(pSrc[i] + pMCELx[i] - pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
    pMCELx += iStrideMCELx;
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr(uiPartAddr)+ (iBlkY-(pSPos->getVer()>>1))*iStrideDst + (iBlkX-(pSPos->getHor()>>1));
  pMCELx = m_pEnhPred[1]->getCrAddr();
  pMCUpBLx = m_pEnhPred[0]->getCrAddr();
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ClipC(pSrc[i] + ((iWeight*(pMCELx[i] - pMCUpBLx[i])+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = ClipC(pSrc[i] + pMCELx[i] - pMCUpBLx[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
    pMCUpBLx += iStrideMCUpBLx;
    pMCELx += iStrideMCELx;
  }
}
#endif
Void TComPrediction::enhPicWithTMWeight(TComPic *pcPicSrc, TComPic *pcPicDst
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSym
#endif
    )
{
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  UInt uiPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  UInt uiPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();
  Int iPUWidth, iPUHeight;

  getProcessUnitSize(iPUWidth, iPUHeight);
  TComMv cBlkPos, cBlkSize;

  for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += g_uiMaxCUHeight)
  {
    Int iBlkHeight = min((Int)g_uiMaxCUHeight, (Int)(uiPicHeight - iBlkY));
    for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += g_uiMaxCUWidth)
    {
      Int iBlkWidth = min((Int)g_uiMaxCUWidth, (Int)(uiPicWidth - iBlkX));              
      //one LCU;
      Int iSubBlkHeight;
      for(Int iSubBlkY=iBlkY; iSubBlkY<iBlkY+iBlkHeight; iSubBlkY += iSubBlkHeight)
      {
        iSubBlkHeight = min(iPUHeight-(iSubBlkY%iPUHeight), (Int)(iBlkY+iBlkHeight - iSubBlkY));
        Int iSubBlkWidth;
        for(Int iSubBlkX =iBlkX; iSubBlkX < iBlkX+iBlkWidth; iSubBlkX += iSubBlkWidth)
        {
          iSubBlkWidth = min(iPUWidth-(iSubBlkX%iPUWidth), (Int)(iBlkX+iBlkWidth - iSubBlkX));              
          
          cBlkPos.set(iSubBlkX, iSubBlkY);
          cBlkSize.set(iSubBlkWidth, iSubBlkHeight);

          getEnhBlockMPUWeight(pcPicSrc, pcPicDst, &cBlkPos, &cBlkSize
#if ILR_TM_ENH_UCM
            , pcPicSym
#endif
            );      
        }
      }//end LCU;
    }
  }
}

Void TComPrediction::getEnhBlockMPUWeight(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
  , TComPicSym *pcPicSymBL
#endif
  )      
{
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
  TComDataCU *pcBLCU;
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  TComPicYuv *pcPicYuvDst = pcPicDst->getPicYuvRec();
  Int iRefIdx[4][2];
  TComMv cMvArr[4][2];  
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  Int iBlkOffset[4][4] = { {0, -1, -1, -1}, {0, 2, -1, -1}, {0, 1, -1, -1}, {0, 1, 2, 3}};
  Int iSubPartSizeX2, iSubPartSizeY2;
  Int iSubPartSizeX, iSubPartSizeY;

  getProcessUnitSize(iSubPartSizeX2, iSubPartSizeY2);
  iSubPartSizeX = iSubPartSizeX2>>1;
  iSubPartSizeY = iSubPartSizeY2>>1;

#if ILR_TM_ENH_UCM
  Int iCurrLayerId = pcPicSrc->getLayerId();
#if ILR_TM_ENH_UCM>=2
  pcPicDst->getCU(0)->getBaseColCU((iBlkX/(4*m_iSRNum[0]))*(4*m_iSRNum[0]), (iBlkY/(4*m_iSRNum[1]))*(4*m_iSRNum[1]), uiBaseCUAddr, uiBaseAbsPartIdx);
#else
  pcPicDst->getCU(0)->getBaseColCU(iBlkX/iSubPartSizeX2*iSubPartSizeX2, iBlkY/iSubPartSizeY2*iSubPartSizeY2, uiBaseCUAddr, uiBaseAbsPartIdx);
#endif
  pcBLCU = pcPicSymBL->getCU(uiBaseCUAddr);
  for(Int i=0; i<4; i++)
    setRefAndMv(pcBLCU, uiBaseAbsPartIdx +i, iRefIdx[i], cMvArr[i]);
#else
  pcPicDst->deriveUnitIdx(iBlkX, iBlkY, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcPicDst->getCU(uiBaseCUAddr);
  //pcBLCU = pcPicSrc->getCU(0)->getBaseColCU(iBlkX, iBlkY, uiBaseCUAddr, uiBaseAbsPartIdx);
  setRefAndMv(pcBLCU, uiBaseAbsPartIdx, iRefIdx[0], cMvArr[0]);
#endif

#if !ILR_TM_ENH_UCM || ILR_TM_ENH_UCM>=2
  Int iPartNum=1;
  PartSize ePartSize = SIZE_2Nx2N;
#else
  Int iPartNum;
  PartSize ePartSize;
  derivePartSize(iRefIdx, cMvArr, iPartNum, ePartSize);
#endif

  for(Int iSubPartIdx=0; iSubPartIdx < iPartNum; iSubPartIdx++)
  {
    Int iSubBlkIdx, iRefIdx0, iRefIdx1;    
    Int iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight;    

    iSubBlkIdx = iBlkOffset[(Int)ePartSize][iSubPartIdx];
    iSubPartX = iBlkX-(iBlkX%iSubPartSizeX2) + ((iSubBlkIdx&1)? iSubPartSizeX: 0);
    iSubPartY = iBlkY-(iBlkY%iSubPartSizeY2) + ((iSubBlkIdx<2)? 0:iSubPartSizeY);    
    if(iSubPartX >= iBlkX + iBlkWidth || iSubPartY >= iBlkY + iBlkHeight)
      continue;
    iSubPartWidth = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_2NxN)? (iSubPartSizeX<<1): iSubPartSizeX;
    iSubPartHeight = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_Nx2N)? (iSubPartSizeY<<1): iSubPartSizeY;
    if(iSubPartX+iSubPartWidth <= iBlkX || iSubPartY+iSubPartHeight <= iBlkY)
      continue;
    iSubPartX = max(iBlkX, iSubPartX);
    iSubPartY = max(iBlkY, iSubPartY);    
    iSubPartWidth = min(iSubPartWidth-(iSubPartX%iSubPartWidth), iBlkX+iBlkWidth-iSubPartX);
    iSubPartHeight = min(iSubPartHeight-(iSubPartY%iSubPartHeight), iBlkY+iBlkHeight-iSubPartY);

    iRefIdx0 = iRefIdx[iSubBlkIdx][0];
    iRefIdx1 = iRefIdx[iSubBlkIdx][1];

#if (ILR_TM_ENH_EXPLICIT_MODE==1)
  if( (pcSlice->getILREnhMethod()==ILR_PIC_NOENH) || (pcSlice->getILREnhMethod()==ILR_BLK_ENH && !pcSlice->getEILRBlockMode()[iSubPartY>>(g_aucConvertToBit[EILR_FLAG_BLK_SIZE]+2)][iSubPartX>>(g_aucConvertToBit[EILR_FLAG_BLK_SIZE]+2)]))
  {
    //iRefIdx0 = iRefIdx1 = -1;
    if(pcSlice->getILREnhMethod()==ILR_PIC_NOENH || (iRefIdx0 <0 && iRefIdx1 <0))
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    else
    {
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        //TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }
      if(!bi)
      {
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iUniRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;        
#if ILR_TM_ENH_UCM
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);

        TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iUniRefIdx )->getFullPelBaseRec();
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iUniRefIdx )->getPicYuvRec();
        xPredInterChromaBlkByPosBilinear( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
      }
      else
      {
#if ILR_TM_ENH_UCM
        TComMv cMv = cMvArr[iSubBlkIdx][0].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
        TComMv cMv1 = cMvArr[iSubBlkIdx][1].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv cMv = cMvArr[iSubBlkIdx][0];
        TComMv cMv1 = cMvArr[iSubBlkIdx][1];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);
        pcSlice->clipMv(cMv1, iSubPartX, iSubPartY);
        TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
        TComYuv* pcMbYuv = m_acYuvPred;
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );

        pcMbYuv = m_acYuvPred+1;
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[0]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight, true );
        TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
        TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
     
        pcMbYuv = m_acYuvPred;
        xPredInterChromaBlkByPosBilinear( pcELxRef0, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        pcMbYuv = m_acYuvPred+1;
        xPredInterChromaBlkByPosBilinear( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[1]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight, true );
      }

      copyBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 0);
      calcEnhBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 1);
      calcEnhBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 2);
    }
    continue;
  }
#endif
  
    if(iRefIdx0 >=0 || iRefIdx1 >=0)
    {
      Int iMode, iWeight;
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
      Int iOffset;
#endif
      Int iWeightIdx;
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }
      if(!bi)
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iUniRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;

        iWeightIdx = pcSlice->getWeightIdxUni(list, iUniRefIdx);
        iMode = pcSlice->getModeArr(0)[iWeightIdx];
        iWeight = pcSlice->getWeight(0, iWeightIdx);
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        iOffset = pcSlice->getOffset(0, iWeightIdx);
#endif

        if(iMode == ILR_DIFF_MODE)
        {        
#if ILR_TM_ENH_UCM
          TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
          TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);

        //Up(BLx); 
        TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iUniRefIdx )->getFullPelBaseRec();
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iUniRefIdx )->getPicYuvRec();
        xPredInterLumaBlkByPosBilinear  ( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
        xPredInterChromaBlkByPosBilinear( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
        }
      }
      else
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        iWeightIdx = pcSlice->getWeightIdxBi(iRefIdx0, iRefIdx1);
        iMode = pcSlice->getModeArr(1)[iWeightIdx];
        iWeight = pcSlice->getWeight(1, iWeightIdx);
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        iOffset = pcSlice->getOffset(1, iWeightIdx);
#endif

        if(iMode == ILR_DIFF_MODE)
        {
#if ILR_TM_ENH_UCM
        TComMv cMv0     = cMvArr[iSubBlkIdx][0].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
        TComMv cMv1     = cMvArr[iSubBlkIdx][1].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv cMv0     = cMvArr[iSubBlkIdx][0];
        TComMv cMv1     = cMvArr[iSubBlkIdx][1];
#endif
        pcSlice->clipMv(cMv0, iSubPartX, iSubPartY);
        pcSlice->clipMv(cMv1, iSubPartX, iSubPartY);

        TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec(); 

        //Up(BLx); 
        TComYuv* pcMbYuv = m_acYuvPred;
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        pcMbYuv = m_acYuvPred+1;
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[0]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight );

          TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
          TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();

          pcMbYuv = m_acYuvPred;
          xPredInterLumaBlkByPosBilinear  ( pcELxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
          xPredInterChromaBlkByPosBilinear( pcELxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
          pcMbYuv = m_acYuvPred+1;
          xPredInterLumaBlkByPosBilinear  ( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
          xPredInterChromaBlkByPosBilinear( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
          m_pEnhPred[1]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight );
        }
      }

      //calc;
      if(iMode == ILR_DIFF_MODE)
      {
        calcEnhBlkModeDiff(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, iWeight
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
                           ,iOffset
#endif
          );
      }
      else
      {
        //assert(!"Not supported yet!");
        printf("iMode:(%d) is not supported yet, treated as intra mode", iMode);
        copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
      }
    }
    else
    {
      //pcPicYuvUpBL -> pcPicYuvDst
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    }
  }
}
#if DEC_EILR_BLK
Void TComPrediction::getEnhBlockMPUWeightDec(TComPic *pcPicSrc, TComPic *pcEILRPic, TComYuv *&pcYuvDst, UInt uiPartAddr, TComMv *pcStartPos, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
  , TComPicSym *pcPicSymBL
#endif
  )      
{
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
  TComDataCU *pcBLCU;
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  //TComPicYuv *pcPicYuvDst = pcEILRPic->getPicYuvRec();
  Int iRefIdx[4][2];
  TComMv cMvArr[4][2];  
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  Int iBlkOffset[4][4] = { {0, -1, -1, -1}, {0, 2, -1, -1}, {0, 1, -1, -1}, {0, 1, 2, 3}};
  Int iSubPartSizeX2, iSubPartSizeY2;
  Int iSubPartSizeX, iSubPartSizeY;

  getProcessUnitSize(iSubPartSizeX2, iSubPartSizeY2);
  iSubPartSizeX = iSubPartSizeX2>>1;
  iSubPartSizeY = iSubPartSizeY2>>1;

#if ILR_TM_ENH_UCM
  Int iCurrLayerId = pcPicSrc->getLayerId();
#if ILR_TM_ENH_UCM>=2
  pcEILRPic->getCU(0)->getBaseColCU((iBlkX/(4*m_iSRNum[0]))*(4*m_iSRNum[0]), (iBlkY/(4*m_iSRNum[1]))*(4*m_iSRNum[1]), uiBaseCUAddr, uiBaseAbsPartIdx);
#else
  pcEILRPic->getCU(0)->getBaseColCU(iBlkX/iSubPartSizeX2*iSubPartSizeX2, iBlkY/iSubPartSizeY2*iSubPartSizeY2, uiBaseCUAddr, uiBaseAbsPartIdx);
#endif
  pcBLCU = pcPicSymBL->getCU(uiBaseCUAddr);
  for(Int i=0; i<4; i++)
    setRefAndMv(pcBLCU, uiBaseAbsPartIdx+i, iRefIdx[i], cMvArr[i]);
#else
  pcEILRPic->deriveUnitIdx(iBlkX, iBlkY, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcEILRPic->getCU(uiBaseCUAddr);
  setRefAndMv(pcBLCU, uiBaseAbsPartIdx, iRefIdx[0], cMvArr[0]);
#endif

#if !ILR_TM_ENH_UCM || ILR_TM_ENH_UCM>=2
  Int iPartNum=1;
  PartSize ePartSize = SIZE_2Nx2N;
#else
  Int iPartNum;
  PartSize ePartSize;
  derivePartSize(iRefIdx, cMvArr, iPartNum, ePartSize);
#endif

  for(Int iSubPartIdx=0; iSubPartIdx < iPartNum; iSubPartIdx++)
  {
    Int iSubBlkIdx, iRefIdx0, iRefIdx1;    
    Int iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight;    

    iSubBlkIdx = iBlkOffset[(Int)ePartSize][iSubPartIdx];
    iSubPartX = iBlkX-(iBlkX%iSubPartSizeX2) + ((iSubBlkIdx&1)? iSubPartSizeX: 0);
    iSubPartY = iBlkY-(iBlkY%iSubPartSizeY2) + ((iSubBlkIdx<2)? 0:iSubPartSizeY);    
    if(iSubPartX >= iBlkX + iBlkWidth || iSubPartY >= iBlkY + iBlkHeight)
      continue;
    iSubPartWidth = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_2NxN)? (iSubPartSizeX<<1): iSubPartSizeX;
    iSubPartHeight = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_Nx2N)? (iSubPartSizeY<<1): iSubPartSizeY;
    if(iSubPartX+iSubPartWidth <= iBlkX || iSubPartY+iSubPartHeight <= iBlkY)
      continue;
    iSubPartX = max(iBlkX, iSubPartX);
    iSubPartY = max(iBlkY, iSubPartY);    
    iSubPartWidth = min(iSubPartWidth-(iSubPartX%iSubPartWidth), iBlkX+iBlkWidth-iSubPartX);
    iSubPartHeight = min(iSubPartHeight-(iSubPartY%iSubPartHeight), iBlkY+iBlkHeight-iSubPartY);

    iRefIdx0 = iRefIdx[iSubBlkIdx][0];
    iRefIdx1 = iRefIdx[iSubBlkIdx][1];

#if (ILR_TM_ENH_EXPLICIT_MODE==1)
  if( (pcSlice->getILREnhMethod()==ILR_PIC_NOENH) || (pcSlice->getILREnhMethod()==ILR_BLK_ENH && !pcSlice->getEILRBlockMode()[iSubPartY>>(g_aucConvertToBit[EILR_FLAG_BLK_SIZE]+2)][iSubPartX>>(g_aucConvertToBit[EILR_FLAG_BLK_SIZE]+2)]))
  {
    //iRefIdx0 = iRefIdx1 = -1;
    if(pcSlice->getILREnhMethod()==ILR_PIC_NOENH || (iRefIdx0 <0 && iRefIdx1 <0))
      copyBlkDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    else
    {
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        //TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }
      if(!bi)
      {
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iUniRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;        
#if ILR_TM_ENH_UCM
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);

        TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iUniRefIdx )->getFullPelBaseRec();
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iUniRefIdx )->getPicYuvRec();
        xPredInterChromaBlkByPosBilinear( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
      }
      else
      {
#if ILR_TM_ENH_UCM
        TComMv cMv = cMvArr[iSubBlkIdx][0].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
        TComMv cMv1 = cMvArr[iSubBlkIdx][1].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv cMv = cMvArr[iSubBlkIdx][0];
        TComMv cMv1 = cMvArr[iSubBlkIdx][1];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);
        pcSlice->clipMv(cMv1, iSubPartX, iSubPartY);
        TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
        //TComYuv* pcMbYuv = m_acYuvPred;
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[0], bi );

        //pcMbYuv = m_acYuvPred+1;
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[1], bi );
        m_pEnhPred[0]->addAvgComp( m_acYuvPred4ELR[0], m_acYuvPred4ELR[1], 0, iSubPartWidth, iSubPartHeight, true );
        TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
        TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
     
        //pcMbYuv = m_acYuvPred;
        xPredInterChromaBlkByPosBilinear( pcELxRef0, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[0], bi );
        //pcMbYuv = m_acYuvPred+1;
        xPredInterChromaBlkByPosBilinear( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[1], bi );
        m_pEnhPred[1]->addAvgComp( m_acYuvPred4ELR[0], m_acYuvPred4ELR[1], 0, iSubPartWidth, iSubPartHeight, true );
      }

      copyBlkOneCompDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 0);
      calcEnhBlkOneCompDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 1);
      calcEnhBlkOneCompDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 2);
    }
    continue;
  }
#endif
  
    if(iRefIdx0 >=0 || iRefIdx1 >=0)
    {
      Int iMode, iWeight;
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
      Int iOffset;
#endif
      Int iWeightIdx;
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }
      if(!bi)
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iUniRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;

        iWeightIdx = pcSlice->getWeightIdxUni(list, iUniRefIdx);
        iMode = pcSlice->getModeArr(0)[iWeightIdx];
        iWeight = pcSlice->getWeight(0, iWeightIdx);
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        iOffset = pcSlice->getOffset(0, iWeightIdx);
#endif
        if(iMode == ILR_DIFF_MODE)
        {        
#if ILR_TM_ENH_UCM
          TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
          TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);

        //Up(BLx); 
        TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iUniRefIdx )->getFullPelBaseRec();
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iUniRefIdx )->getPicYuvRec();
        xPredInterLumaBlkByPosBilinear  ( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
        xPredInterChromaBlkByPosBilinear( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
        }
      }
      else
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        iWeightIdx = pcSlice->getWeightIdxBi(iRefIdx0, iRefIdx1);
        iMode = pcSlice->getModeArr(1)[iWeightIdx];
        iWeight = pcSlice->getWeight(1, iWeightIdx);
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
        iOffset = pcSlice->getOffset(1, iWeightIdx);
#endif

        if(iMode == ILR_DIFF_MODE)
        {
#if ILR_TM_ENH_UCM
        TComMv cMv0     = cMvArr[iSubBlkIdx][0].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
        TComMv cMv1     = cMvArr[iSubBlkIdx][1].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv cMv0     = cMvArr[iSubBlkIdx][0];
        TComMv cMv1     = cMvArr[iSubBlkIdx][1];
#endif
        pcSlice->clipMv(cMv0, iSubPartX, iSubPartY);
        pcSlice->clipMv(cMv1, iSubPartX, iSubPartY);

        TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec(); 

        //Up(BLx); 
        //TComYuv* pcMbYuv = m_acYuvPred;
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[0], bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[0], bi );
        //pcMbYuv = m_acYuvPred+1;
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[1], bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[1], bi );
        m_pEnhPred[0]->addAvgComp( m_acYuvPred4ELR[0], m_acYuvPred4ELR[1], 0, iSubPartWidth, iSubPartHeight );

          TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
          TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();

          //pcMbYuv = m_acYuvPred;
          xPredInterLumaBlkByPosBilinear  ( pcELxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[0], bi );
          xPredInterChromaBlkByPosBilinear( pcELxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[0], bi );
          //pcMbYuv = m_acYuvPred+1;
          xPredInterLumaBlkByPosBilinear  ( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[1], bi );
          xPredInterChromaBlkByPosBilinear( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, m_acYuvPred4ELR[1], bi );
          m_pEnhPred[1]->addAvgComp( m_acYuvPred4ELR[0], m_acYuvPred4ELR[1], 0, iSubPartWidth, iSubPartHeight );
        }
      }

      //calc;
      if(iMode == ILR_DIFF_MODE)
      {
          calcEnhBlkModeDiffDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, iWeight
#if ILR_TM_ENH_WEIGHT_SIGNAL==2
            ,iOffset
#endif
            );
      }
      else
      {
        //assert(!"Not supported yet!");
        printf("iMode:(%d) is not supported yet, treated as intra mode", iMode);
        copyBlkDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
      }
    }
    else
    {
      //pcPicYuvUpBL -> pcPicYuvDst
      copyBlkDec(pcPicYuvUpBL, pcYuvDst, uiPartAddr, pcStartPos, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    }
  }
}
#endif

//the differential or residual signal is ready stored in pcPicDst;
Void TComPrediction::enhPicWithTMWeightDiff(TComPic *pcPicSrc, TComPic *pcPicDst
#if ILR_TM_ENH_UCM
    , TComPicSym *pcPicSym
#endif
    )
{
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  UInt uiPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  UInt uiPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();
  Int iPUWidth, iPUHeight;
  TComMv cBlkPos, cBlkSize;

  getProcessUnitSize(iPUWidth, iPUHeight);
  for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += g_uiMaxCUHeight)
  {
    Int iBlkHeight = min((Int)g_uiMaxCUHeight, (Int)(uiPicHeight - iBlkY));
    for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += g_uiMaxCUWidth)
    {
      Int iBlkWidth = min((Int)g_uiMaxCUWidth, (Int)(uiPicWidth - iBlkX));              
      //one LCU;
      Int iSubBlkHeight;
      for(Int iSubBlkY=iBlkY; iSubBlkY<iBlkY+iBlkHeight; iSubBlkY += iSubBlkHeight)
      {
        iSubBlkHeight = min(iPUHeight-(iSubBlkY%iPUHeight), (Int)(iBlkY+iBlkHeight - iSubBlkY));
        Int iSubBlkWidth;
        for(Int iSubBlkX =iBlkX; iSubBlkX < iBlkX+iBlkWidth; iSubBlkX += iSubBlkWidth)
        {
          iSubBlkWidth = min(iPUWidth-(iSubBlkX%iPUWidth), (Int)(iBlkX+iBlkWidth - iSubBlkX));              
          
          cBlkPos.set(iSubBlkX, iSubBlkY);
          cBlkSize.set(iSubBlkWidth, iSubBlkHeight);

          getEnhBlockMPUWeight_D(pcPicSrc, pcPicDst, &cBlkPos, &cBlkSize
#if ILR_TM_ENH_UCM
            , pcPicSym
#endif
            );      
        }
      }//end LCU;
    }
  }
}

Void TComPrediction::getEnhBlockMPUWeight_D(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
  , TComPicSym *pcPicSymBL
#endif
  )      
{
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
  TComDataCU *pcBLCU;
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  TComPicYuv *pcPicYuvDst = pcPicDst->getPicYuvRec();
  Int iRefIdx[4][2];
  TComMv cMvArr[4][2];  
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  Int iBlkOffset[4][4] = { {0, -1, -1, -1}, {0, 2, -1, -1}, {0, 1, -1, -1}, {0, 1, 2, 3}};
  Int iSubPartSizeX2, iSubPartSizeY2;
  Int iSubPartSizeX, iSubPartSizeY;

  getProcessUnitSize(iSubPartSizeX2, iSubPartSizeY2);
  iSubPartSizeX = iSubPartSizeX2>>1;
  iSubPartSizeY = iSubPartSizeY2>>1;

#if ILR_TM_ENH_UCM
#if ILR_TM_ENH_UCM>=2
  pcPicDst->getCU(0)->getBaseColCU((iBlkX/(4*m_iSRNum[0]))*(4*m_iSRNum[0]), (iBlkY/(4*m_iSRNum[1]))*(4*m_iSRNum[1]), uiBaseCUAddr, uiBaseAbsPartIdx);
#else
  pcPicDst->getCU(0)->getBaseColCU(iBlkX/iSubPartSizeX2*iSubPartSizeX2, iBlkY/iSubPartSizeY2*iSubPartSizeY2, uiBaseCUAddr, uiBaseAbsPartIdx);
#endif
  pcBLCU = pcPicSymBL->getCU(uiBaseCUAddr);
  for(Int i=0; i<4; i++)
    setRefAndMv(pcBLCU, uiBaseAbsPartIdx+i, iRefIdx[i], cMvArr[i]);
#else
  pcPicDst->deriveUnitIdx(iBlkX, iBlkY, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcPicDst->getCU(uiBaseCUAddr);
  setRefAndMv(pcBLCU, uiBaseAbsPartIdx, iRefIdx[0], cMvArr[0]);
#endif

#if !ILR_TM_ENH_UCM || ILR_TM_ENH_UCM>=2
  Int iPartNum=1;
  PartSize ePartSize = SIZE_2Nx2N;
#else
  Int iPartNum;
  PartSize ePartSize;
  derivePartSize(iRefIdx, cMvArr, iPartNum, ePartSize);
#endif

  for(Int iSubPartIdx=0; iSubPartIdx < iPartNum; iSubPartIdx++)
  {
    Int iSubBlkIdx, iRefIdx0, iRefIdx1;    
    Int iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight;    

    iSubBlkIdx = iBlkOffset[(Int)ePartSize][iSubPartIdx];
    iSubPartX = iBlkX-(iBlkX%iSubPartSizeX2) + ((iSubBlkIdx&1)? iSubPartSizeX: 0);
    iSubPartY = iBlkY-(iBlkY%iSubPartSizeY2) + ((iSubBlkIdx<2)? 0:iSubPartSizeY);    
    if(iSubPartX >= iBlkX + iBlkWidth || iSubPartY >= iBlkY + iBlkHeight)
      continue;
    iSubPartWidth = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_2NxN)? (iSubPartSizeX<<1): iSubPartSizeX;
    iSubPartHeight = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_Nx2N)? (iSubPartSizeY<<1): iSubPartSizeY;
    if(iSubPartX+iSubPartWidth <= iBlkX || iSubPartY+iSubPartHeight <= iBlkY)
      continue;
    iSubPartX = max(iBlkX, iSubPartX);
    iSubPartY = max(iBlkY, iSubPartY);    
    iSubPartWidth = min(iSubPartWidth-(iSubPartX%iSubPartWidth), iBlkX+iBlkWidth-iSubPartX);
    iSubPartHeight = min(iSubPartHeight-(iSubPartY%iSubPartHeight), iBlkY+iBlkHeight-iSubPartY);

    iRefIdx0 = iRefIdx[iSubBlkIdx][0];
    iRefIdx1 = iRefIdx[iSubBlkIdx][1];

#if (ILR_TM_ENH_EXPLICIT_MODE==1)
    if( (pcSlice->getILREnhMethod()==ILR_PIC_NOENH) || (pcSlice->getILREnhMethod()==ILR_BLK_ENH && !pcSlice->getEILRBlockMode()[iSubPartY>>(g_aucConvertToBit[EILR_FLAG_BLK_SIZE]+2)][iSubPartX>>(g_aucConvertToBit[EILR_FLAG_BLK_SIZE]+2)]))
    {
      //iRefIdx0 = iRefIdx1 = -1;
      if(pcSlice->getILREnhMethod()==ILR_PIC_NOENH || (iRefIdx0 <0 && iRefIdx1 <0))
      {
        copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
      }
      else
      {
       copyBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 0);
       calcEnhBlk2OneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 1);
       calcEnhBlk2OneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 2);
      }
      continue;
    }
#endif
  
    if(iRefIdx0 >=0 || iRefIdx1 >=0)
    {
      Int iMode, iWeight;
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
      Int iOffset;
#endif
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }

      if(!bi)
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iUniRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;
        Int iWeightIdx = pcSlice->getWeightIdxUni(list, iUniRefIdx);
        iMode = pcSlice->getModeArr(0)[iWeightIdx];
        iWeight = pcSlice->getWeight(0, iWeightIdx);
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
        iOffset = pcSlice->getOffset(0, iWeightIdx);
#endif
      }
      else
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        Int iWeightIdx = pcSlice->getWeightIdxBi(iRefIdx0, iRefIdx1);
        iMode = pcSlice->getModeArr(1)[iWeightIdx];
        iWeight = pcSlice->getWeight(1, iWeightIdx);
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
        iOffset = pcSlice->getOffset(1, iWeightIdx);
#endif
      }

      //calc;
      if(iMode == ILR_DIFF_MODE)
      {
        //if(iWeight)
          calcEnhBlkModeDiff2(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, iWeight
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
                             ,iOffset
#endif
            );
        //else
          //calcEnhBlk2(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
      }
      else
      {
        //assert(!"Not supported yet!");
        printf("iMode:(%d) is not supported yet, treated as intra mode", iMode);
        copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
      }
    }
    else
    {
      //pcPicYuvUpBL -> pcPicYuvDst
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    }
  }
}

/*
Void TComPrediction::calcEnhBlk2(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight) //, UInt uiAbsPartIdx)
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i+=2)
    {
      pDst[i] = ClipY(pSrc[i]+pDst[i]);
      pDst[i+1] = ClipY(pSrc[i+1]+pDst[i+1]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
  }
  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+ pDst[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      pDst[i] = ClipC(pSrc[i]+ pDst[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
  }
}
*/
Void TComPrediction::calcEnhBlkModeDiff2(TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight, Int iWeight
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
                                         ,Int iOffset
#endif
                                        )
{
  //dst = up(BL) + (1-weight)*(MC(ELx)-MC(Up(BLx)));
  Int iStrideSrc = pcPicYuvSrc->getStride();
  Int iStrideDst = pcPicYuvDst->getStride();
  Pel *pSrc = pcPicYuvSrc->getLumaAddr() + iBlkY*iStrideSrc + iBlkX;
  Pel *pDst = pcPicYuvDst->getLumaAddr() + iBlkY*iStrideDst + iBlkX;

  if(iWeight)
  {
    iWeight = WEIGHT_SCALE - iWeight;
    if(iWeight)
    {
      for(Int j=0; j<iBlkHeight; j++)
      {
        for(Int i=0; i<iBlkWidth; i+=2)
        {
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
          pDst[i] = ClipY(pSrc[i]+ ((iWeight*pDst[i] + WEIGHT_SCALE_OFFSET + iOffset)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ ((iWeight*pDst[i+1]+WEIGHT_SCALE_OFFSET + iOffset)>>WEIGHT_SCALE_SHIFT));
#else
          pDst[i] = ClipY(pSrc[i]+ ((iWeight*pDst[i] + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
          pDst[i+1] = ClipY(pSrc[i+1]+ ((iWeight*pDst[i+1]+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
#endif
        }
        pDst += iStrideDst;
        pSrc += iStrideSrc;
      }
    }
    else
    {
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
      iOffset = (iOffset + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT;
#endif
      for(Int j=0; j<iBlkHeight; j++)
      {
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
        for(Int i=0; i<iBlkWidth; i++)
          pDst[i] = ClipY(pSrc[i]+ iOffset);
#else
        memcpy(pDst, pSrc, iBlkWidth*sizeof(Pel));
#endif
        pDst += iStrideDst;
        pSrc += iStrideSrc;
      }
    }
  }
  else
  {
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
    iOffset = (iOffset + WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT;
#endif
    for(Int j=0; j<iBlkHeight; j++)
    {
      for(Int i=0; i<iBlkWidth; i+=2)
      {
#if (ILR_TM_ENH_WEIGHT_SIGNAL==2)
        pDst[i] = ClipY(pSrc[i]+ pDst[i] + iOffset);
        pDst[i+1] = ClipY(pSrc[i+1]+ pDst[i+1] + iOffset);
#else
        pDst[i] = ClipY(pSrc[i]+ pDst[i]);
        pDst[i+1] = ClipY(pSrc[i+1]+ pDst[i+1]);
#endif
      }
      pDst += iStrideDst;
      pSrc += iStrideSrc;
    }
  }

  iStrideSrc = pcPicYuvSrc->getCStride();
  iStrideDst = pcPicYuvDst->getCStride();
  iBlkWidth >>=1;
  iBlkHeight >>=1;
  iBlkX >>= 1;
  iBlkY >>= 1;

  pSrc = pcPicYuvSrc->getCbAddr() + iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCbAddr() + iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ClipC(pSrc[i] + ((iWeight*pDst[i]+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = ClipC(pSrc[i] + pDst[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
  }

  pSrc = pcPicYuvSrc->getCrAddr()+ iBlkY*iStrideSrc + iBlkX;
  pDst = pcPicYuvDst->getCrAddr()+ iBlkY*iStrideDst + iBlkX;
  for(Int j=0; j<iBlkHeight; j++)
  {
    for(Int i=0; i<iBlkWidth; i++)
    {
      //pDst[i] = ClipC(pSrc[i] + ((iWeight*pDst[i]+WEIGHT_SCALE_OFFSET)>>WEIGHT_SCALE_SHIFT));
      pDst[i] = ClipC(pSrc[i] + pDst[i]);
    }
    pDst += iStrideDst;
    pSrc += iStrideSrc;
  }
}

#else

Void TComPrediction::enhPicWithTM(TComPic *pcPicSrc, TComPic *pcPicDst
#if ILR_TM_ENH_UCM
                                  , TComPicSym *pcPicSym
#endif
  )
{
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  UInt uiPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  UInt uiPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();
  Int iPUWidth, iPUHeight;
  getProcessUnitSize(iPUWidth, iPUHeight);

  TComMv cBlkPos, cBlkSize;
  for ( Int iBlkY = 0; iBlkY < uiPicHeight; iBlkY += g_uiMaxCUHeight)
  {
    Int iBlkHeight = min((Int)g_uiMaxCUHeight, (Int)(uiPicHeight - iBlkY));
    for(Int iBlkX =0; iBlkX < uiPicWidth; iBlkX += g_uiMaxCUWidth)
    {
      Int iBlkWidth = min((Int)g_uiMaxCUWidth, (Int)(uiPicWidth - iBlkX));              
      //one LCU;
      Int iSubBlkHeight;
      for(Int iSubBlkY=iBlkY; iSubBlkY<iBlkY+iBlkHeight; iSubBlkY += iSubBlkHeight)
      {
        iSubBlkHeight = min(iPUHeight-(iSubBlkY%iPUHeight), (Int)(iBlkY+iBlkHeight - iSubBlkY));
        Int iSubBlkWidth;
        for(Int iSubBlkX =iBlkX; iSubBlkX < iBlkX+iBlkWidth; iSubBlkX += iSubBlkWidth)
        {
          iSubBlkWidth = min(iPUWidth-(iSubBlkX%iPUWidth), (Int)(iBlkX+iBlkWidth - iSubBlkX));              
          
          cBlkPos.set(iSubBlkX, iSubBlkY);
          cBlkSize.set(iSubBlkWidth, iSubBlkHeight);

          getEnhBlockMPU(pcPicSrc, pcPicDst, &cBlkPos, &cBlkSize
#if ILR_TM_ENH_UCM
            , pcPicSym
#endif
            );      
        }
      }
    }
  }
}

Void TComPrediction::getEnhBlockMPU(TComPic *pcPicSrc, TComPic *pcPicDst, TComMv *pcBlkPos, TComMv *pcBlkSize
#if ILR_TM_ENH_UCM
                                    , TComPicSym *pcPicSymBL
#endif
  )      
{
  //dst = up(BL) + MC(ELx)-MC(Up(BLx));
  Int iBlkX = pcBlkPos->getHor();
  Int iBlkY = pcBlkPos->getVer();
  Int iBlkWidth = pcBlkSize->getHor();
  Int iBlkHeight = pcBlkSize->getVer();  
  TComSlice *pcSlice = pcPicSrc->getSlice(pcPicSrc->getCurrSliceIdx());
  TComPic *pcPicBL = pcSlice->getBaseColPic();
  TComDataCU *pcBLCU;
  TComPicYuv *pcPicYuvUpBL = pcPicSrc->getFullPelBaseRec();
  TComPicYuv *pcPicYuvDst = pcPicDst->getPicYuvRec();
  Int iRefIdx[4][2];
  TComMv cMvArr[4][2];  
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;  
  Int iBlkOffset[4][4] = { {0, -1, -1, -1}, {0, 2, -1, -1}, {0, 1, -1, -1}, {0, 1, 2, 3}};
  Int iCurrLayerId = pcPicSrc->getLayerId();
  Int iSubPartSizeX2, iSubPartSizeY2;
  Int iSubPartSizeX, iSubPartSizeY;

  getProcessUnitSize(iSubPartSizeX2, iSubPartSizeY2);
  iSubPartSizeX = iSubPartSizeX2>>1;
  iSubPartSizeY = iSubPartSizeY2>>1;

#if ILR_TM_ENH_UCM
#if ILR_TM_ENH_UCM>=2
  pcPicDst->getCU(0)->getBaseColCU((iBlkX/(4*m_iSRNum[0]))*(4*m_iSRNum[0]), (iBlkY/(4*m_iSRNum[1]))*(4*m_iSRNum[1]), uiBaseCUAddr, uiBaseAbsPartIdx);
#else
  pcPicDst->getCU(0)->getBaseColCU(iBlkX/iSubPartSizeX2*iSubPartSizeX2, iBlkY/iSubPartSizeY2*iSubPartSizeY2, uiBaseCUAddr, uiBaseAbsPartIdx);
#endif
  pcBLCU = pcPicSymBL->getCU(uiBaseCUAddr);
  for(Int i=0; i<4; i++)
    setRefAndMv(pcBLCU, uiBaseAbsPartIdx +i, iRefIdx[i], cMvArr[i]);
#else
  pcPicDst->deriveUnitIdx(iBlkX, iBlkY, uiBaseCUAddr, uiBaseAbsPartIdx);
  pcBLCU = pcPicDst->getCU(uiBaseCUAddr);
  setRefAndMv(pcBLCU, uiBaseAbsPartIdx, iRefIdx[0], cMvArr[0]);
#endif

#if !ILR_TM_ENH_UCM || ILR_TM_ENH_UCM>=2
  Int iPartNum=1;
  PartSize ePartSize = SIZE_2Nx2N;
#else
  Int iPartNum;
  PartSize ePartSize;
  derivePartSize(iRefIdx, cMvArr, iPartNum, ePartSize);
#endif

  for(Int iSubPartIdx=0; iSubPartIdx < iPartNum; iSubPartIdx++)
  {
    Int iSubBlkIdx, iRefIdx0, iRefIdx1;    
    Int iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight;
    
    iSubBlkIdx = iBlkOffset[(Int)ePartSize][iSubPartIdx];
    iSubPartX = iBlkX-(iBlkX%iSubPartSizeX2) + ((iSubBlkIdx&1)? iSubPartSizeX: 0);
    iSubPartY = iBlkY-(iBlkY%iSubPartSizeY2) + ((iSubBlkIdx<2)? 0:iSubPartSizeY);    
    if(iSubPartX >= iBlkX + iBlkWidth || iSubPartY >= iBlkY + iBlkHeight)
      continue;
    iSubPartWidth = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_2NxN)? (iSubPartSizeX<<1): iSubPartSizeX;
    iSubPartHeight = (ePartSize == SIZE_2Nx2N || ePartSize == SIZE_Nx2N)? (iSubPartSizeY<<1): iSubPartSizeY;
    if(iSubPartX+iSubPartWidth <= iBlkX || iSubPartY+iSubPartHeight <= iBlkY)
      continue;
    iSubPartX = max(iBlkX, iSubPartX);
    iSubPartY = max(iBlkY, iSubPartY);    
    iSubPartWidth = min(iSubPartWidth-(iSubPartX%iSubPartWidth), iBlkX+iBlkWidth-iSubPartX);
    iSubPartHeight = min(iSubPartHeight-(iSubPartY%iSubPartHeight), iBlkY+iBlkHeight-iSubPartY);

    iRefIdx0 = iRefIdx[iSubBlkIdx][0];
    iRefIdx1 = iRefIdx[iSubBlkIdx][1];

#if (ILR_TM_ENH_EXPLICIT_MODE==1)
  if( (pcSlice->getILREnhMethod()==ILR_PIC_NOENH) || (pcSlice->getILREnhMethod()==ILR_BLK_ENH && !pcSlice->getEILRBlockMode()[iSubPartY>>(g_aucConvertToBit[g_uiMaxCUHeight]+2)][iSubPartX>>(g_aucConvertToBit[g_uiMaxCUWidth]+2)]))
  {
    //iRefIdx0 = iRefIdx1 = -1;
    if(pcSlice->getILREnhMethod()==ILR_PIC_NOENH || (iRefIdx0 <0 && iRefIdx1 <0))
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    else
    {
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }
      if(!bi)
      {
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;        
#if ILR_TM_ENH_UCM
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);
        TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iRefIdx )->getFullPelBaseRec();
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );

        TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iRefIdx )->getPicYuvRec();
        xPredInterChromaBlkByPosBilinear( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
      }
      else
      {
#if ILR_TM_ENH_UCM
        TComMv cMv = cMvArr[iSubBlkIdx][0].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
        TComMv cMv1 = cMvArr[iSubBlkIdx][1].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv cMv = cMvArr[iSubBlkIdx][0];
        TComMv cMv1 = cMvArr[iSubBlkIdx][1];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);
        pcSlice->clipMv(cMv1, iSubPartX, iSubPartY);
        TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
        TComYuv* pcMbYuv = m_acYuvPred;
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );

        pcMbYuv = m_acYuvPred+1;
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[0]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight, true );

        TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
        TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
     
        pcMbYuv = m_acYuvPred;
        xPredInterChromaBlkByPosBilinear( pcELxRef0, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        pcMbYuv = m_acYuvPred+1;
        xPredInterChromaBlkByPosBilinear( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[1]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight, true );
      }

      copyBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 0);
      calcEnhBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 1);
      calcEnhBlkOneComp(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 2);
    }
    continue;
  }
#endif

    if(iRefIdx0 >=0 || iRefIdx1 >=0)
    {
      Bool bi = (iRefIdx0 >=0 && iRefIdx1 >=0);
      if(bi)
      {
        TComSlice *pcSliceBL = pcPicBL->getSlice(pcPicBL->getCurrSliceIdx());
        Int RefPOCL0 = pcSliceBL->getRefPic(REF_PIC_LIST_0, iRefIdx0)->getPOC();
        Int RefPOCL1 = pcSliceBL->getRefPic(REF_PIC_LIST_1, iRefIdx1)->getPOC();
        if(RefPOCL0 == RefPOCL1 && cMvArr[iSubBlkIdx][0] == cMvArr[iSubBlkIdx][1])
        {
          bi = false;
        }
      }
      if(!bi)
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        RefPicList list = iRefIdx0 >=0? REF_PIC_LIST_0 : REF_PIC_LIST_1;
        Int        iUniRefIdx = iRefIdx0 >=0?  iRefIdx0 : iRefIdx1;
        
        TComPicYuv *pcUpBLxRef = pcSlice->getRefPic( list, iUniRefIdx )->getFullPelBaseRec();
        TComPicYuv *pcELxRef = pcSlice->getRefPic( list, iUniRefIdx )->getPicYuvRec();
#if ILR_TM_ENH_UCM
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv     cMv     = cMvArr[iSubBlkIdx][(Int)(list)];
#endif
        pcSlice->clipMv(cMv, iSubPartX, iSubPartY);
        //Up(BLx); 
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[0], bi );
        //ELx
        xPredInterLumaBlkByPosBilinear  ( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
        xPredInterChromaBlkByPosBilinear( pcELxRef, iSubPartX, iSubPartY, &cMv, iSubPartWidth, iSubPartHeight, m_pEnhPred[1], bi );
      }
      else
      {
        //assert(iRefIdx0 < pcSlice->getNumRefIdx(REF_PIC_LIST_0));
        //assert(iRefIdx1 < pcSlice->getNumRefIdx(REF_PIC_LIST_1));
        TComPicYuv *pcUpBLxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        TComPicYuv *pcUpBLxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
        TComPicYuv *pcELxRef0 = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
        TComPicYuv *pcELxRef1 = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
#if ILR_TM_ENH_UCM
        TComMv cMv0     = cMvArr[iSubBlkIdx][0].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
        TComMv cMv1     = cMvArr[iSubBlkIdx][1].scaleMv(g_mvScalingFactor[iCurrLayerId][0], g_mvScalingFactor[iCurrLayerId][1]);
#else
        TComMv cMv0     = cMvArr[iSubBlkIdx][0];
        TComMv cMv1     = cMvArr[iSubBlkIdx][1];
#endif
        pcSlice->clipMv(cMv0, iSubPartX, iSubPartY);
        pcSlice->clipMv(cMv1, iSubPartX, iSubPartY);
        //Up(BLx); 
        TComYuv* pcMbYuv = m_acYuvPred;
        //TComPicYuv *pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getFullPelBaseRec();
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );

        pcMbYuv = m_acYuvPred+1;
        //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getFullPelBaseRec();
        xPredInterLumaBlkByPosBilinear  ( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        xPredInterChromaBlkByPosBilinear( pcUpBLxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[0]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight );
        //ELx
        pcMbYuv = m_acYuvPred;
        //pcRef = pcSlice->getRefPic( REF_PIC_LIST_0, iRefIdx0 )->getPicYuvRec();
        xPredInterLumaBlkByPosBilinear  ( pcELxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        xPredInterChromaBlkByPosBilinear( pcELxRef0, iSubPartX, iSubPartY, &cMv0, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        pcMbYuv = m_acYuvPred+1;
        //pcRef = pcSlice->getRefPic( REF_PIC_LIST_1, iRefIdx1 )->getPicYuvRec();
        xPredInterLumaBlkByPosBilinear  ( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        xPredInterChromaBlkByPosBilinear( pcELxRef1, iSubPartX, iSubPartY, &cMv1, iSubPartWidth, iSubPartHeight, pcMbYuv, bi );
        m_pEnhPred[1]->addAvgComp( m_acYuvPred, m_acYuvPred+1, 0, iSubPartWidth, iSubPartHeight );
      }
      //calc;
      //calcEnhBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
      calcEnhBlkModeDiff(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight, 0);
    }
    else
    {
      //pcPicYuvUpBL -> pcPicYuvDst
      copyBlk(pcPicYuvUpBL, pcPicYuvDst, iSubPartX, iSubPartY, iSubPartWidth, iSubPartHeight);
    }
  }
}

#endif


#endif  //ILR_TM_ENH
//! \}
